#ifndef __LINKFACE_SDK_V2_H__
#define __LINKFACE_SDK_V2_H__

#include "list.h"

#ifdef __cplusplus
extern "C"
{
#endif

#define LENGTH_PERSON_ID            65
#define LENGTH_PERSON_NAME          128
#define LENGTH_FACE_MD5             33
#define LENGTH_GROUP_ID             64
#define LENGTH_EVENTLIST            1024  // 事件队列长度
#define MAX_COUNT_GROUP_ID          1024
#define MAX_LENGTH_EXTRA_INFO       2048

/* 错误码枚举值 */
typedef enum {
    LF_ERR_SUCCESS = 0,                   //成功，无错

    // 添加人脸底库错误码表 
    LF_ERR_IMAGE_PARSEPIC_FAIL = 16001, //图片解析失败
    LF_ERR_IMAGE_PIC_TOOBIG = 16002, //图片过大
    LF_ERR_IMAGE_PIC_TOOSMALL = 16003, //图片过小
    LF_ERR_IMAGE_NO_FACE = 16004, //图片中未检测到人脸
    LF_ERR_IMAGE_MULTI_FACES = 16005, //图片中检测到多人脸
    LF_ERR_IMAGE_FACE_EXIST = 16006, //人脸已存在
    LF_ERR_IMAGE_PIC_FORMATE_ERR = 16007, //不支持的图片格式
    LF_ERR_IMAGE_PIX_ERR = 16008, //不支持的像素
    LF_ERR_IMAGE_UNKNOWN_ERR = 16009, //其他未知错误
    LF_ERR_IMAGE_ALG_ERR = 16010, //算法异常，无法提特征
    LF_ERR_IMAGE_FACE_DB_FULL = 16011, //人脸底库达到最大值

    // 文件下载错误码列表
    LF_ERR_URL_DOWNLOAD_FAILED = 17001
} LINKFACE_ERROR_CODE_E;

/*新增人脸底库需要的字段*/
typedef struct {
    /*! 人员ID: 可能是身份证号码，也可能是其他身份唯一标示。 */
    char person_id[LENGTH_PERSON_ID];

    /*! 人脸数据md5值。*/
    char faceimg_md5[LENGTH_FACE_MD5];

    /*! 人脸图片数据大小。单位：字节。仅在注册人脸时有效。 */
    int faceimg_size;

    /*! 人脸图像原始数据。*/
    char *faceimg;

    /*! 新增本次人脸数据时的时间戳, 单位为ms。*/
    long long timestamp_ms;
} linkface_image_info;

/* 
 * 人脸照片添加队列
 */
typedef struct __tag_add_face_image_list {
    struct list_head face_list;
    linkface_image_info face_image_info;
} LINKFACE_ADD_FACE_IMAGE_LIST_ST;

typedef struct {
    /*! 人员ID: 可能是身份证号码，也可能是其他身份唯一标示。 */
    char person_id[LENGTH_PERSON_ID];

    /*! 人员姓名，可能为空。*/
    char name[LENGTH_PERSON_NAME];

    /*! 该人员所在的组ID，一个人可能存在于多个组中*/
    char group_ids[MAX_COUNT_GROUP_ID][LENGTH_GROUP_ID];

    int group_id_count;

    /*! 新增该人员信息时的时间戳，单位为ms。*/
    long long timestamp_ms;

    /*! 注册人脸时的其他信息。 */
    char *extra_info;
} linkface_user_info;

/// 人脸比对状态。
typedef enum {
    /*! 比对失败。 */
    LINKFACE_MATCH_STATUS_FAILED = -1,

    /*! 未进行比对。 */
    LINKFACE_MATCH_STATUS_NULL = 0,

    /*! 比对成功。 */
    LINKFACE_MATCH_STATUS_SUCCESS = 1,
} linkface_match_status_t;

// 增量事件类型
typedef enum {
    //用户信息
    LINKFACE_SYNC_EVENT_USERINFO = 1,
    //用户图片
    LINKFACE_SYNC_EVENT_USERPIC = 2,
    //用户特征
    LINKFACE_SYNC_EVENT_USERFEATURE = 3,
    //用户快速删除
    LINKFACE_SYNC_EVENT_USERFASTDEL = 4,
    //删除用户组
    LINKFACE_SYNC_EVENT_GROUPDEL = 5,
    //清空所有数据
    LINKFACE_SYNC_EVENT_DELTOTAL = 21
} LINKFACE_SYNC_EVENT_E;

/*人脸对比上报*/
/*注册人脸需要的字段*/
typedef struct {
    /*! 比对状态。 */
    linkface_match_status_t match_status; //对比成功才会有人脸(即person_id，match_score)的值才有意义

    /*! 人脸比对到的人员ID。 抓拍成功时，必要字段。*/
    char person_id[64];

    /*! 对比相似度。 抓拍成功时，必要字段。范围为[1, 100]*/
    int match_score;

    /*! 人脸全景图大小。单位为Byte，不能超过10MByte. */
    int panorama_img_size; //大小大于0时全景图的数据才有意义，否则全景图所有数据全部置空

    /*! 人脸全景图原始数据。必要字段。 */
    char *panorama_img;

    /*! 人脸全景图名称。 必要字段。 */
    char *panorama_img_name;

    /*! 事件产生utc时间。 必要字段。 */
    long long timestamp_ms;
} linkface_match_result;

/* 
 * 事件上报队列
 */
typedef struct __tag_event_list {
    struct list_head event_list;
    linkface_match_result match_result;
} LINKFACE_EVENT_LIST_ST;

/** 云端下发底库同步请求时，sdk先缓存人脸图片，再调用该接口将图片批量入库。
 * @brief       批量注册人脸数据接收回调函数。
 * @param[IN]   pImageInfoList，图片信息列表。
 * @param[IN]   num，该值为pImageInfoList、pErrorCode的个数。
 * @param[OUT]  pErrorCode，每张图片的入库结果
 * @ret         void
 *
 * 注意:
 * 1. 该接口为同步调用，接口不宜耗时过久。
 * 2. 该接口不会在多线程中并发调用。
 * 3. 每张图片的添加结果，都要返回到pErrorCode中
 * 4. 错误码请严格按照 @LINKFACE_ERROR_CODE_IMAGE_E 中的定义. 
 */
typedef int (*linkface_add_image_batch_cb)(const struct list_head *pImageInfoList, int num, int *pErrorCode);

/* 
 * 添加用户信息，成功返回0，失败返回其他值.
 */
typedef int (*linkface_add_user_cb)(const linkface_user_info *info, void *user);

/** 云端需要删除某个人脸底库时，sdk会调用该接口将人脸信息逐个传递给厂商，厂商需要将该人脸从算法数据库中删除.
 * @brief   删除人脸数据接收回调函数格式定义。
 * @param   人脸id。
 * @param   user 用户参数，用户设置回调时指定的参数通过该参数再次回传给用户。
 * @ret     如果删除人脸成功，则返回0，否者返回其他值。
 *
 * 注意:
 * 1. 只有人脸数据真实存在时(person_id/faceimg_md5组合校验)，才返回成功。
 * 2. person_id 不存在返回-1。 
 * 3. person_id 存在而 faceimg_md5 不存在，返回-2。 
 * */
typedef int (*linkface_del_image_cb)(const linkface_image_info *info, void *user);

/*
 * 删除用户信息, 成功返回0，失败返回其他值.
 * */
typedef int (*linkface_del_user_cb)(const linkface_user_info *info, void *user);

/*
 * 基于入库的时间戳，轮询人脸底库信息。
 * */
typedef void (*linkface_polling_image_cb)(long long timestamp_ms, void *user,
                                          void (*cb_sdk)(const linkface_image_info *info, void *user), void *user_sdk);

/*
 * 基于入库的时间戳，轮询人员库信息。
 * */
typedef void (*linkface_polling_user_info_cb)(long long timestamp_ms, void *user,
                                              void (*cb_sdk)(const linkface_user_info *info, void *user),
                                              void *user_sdk);

/*
 * 基于人员id，查询人脸底库信息。
 * 厂商实现，结构体内存由sdk释放。
 * */
typedef linkface_image_info *(*linkface_query_image_info_cb)(char *person_id, void *user);

/*
 * 基于人员id，查询人员信息.
 * */
typedef linkface_user_info *(*linkface_query_user_info_cb)(char *person_id, void *user);

/*
 * 查询端侧人脸图片总数
 * */
typedef int (*linkface_querey_image_count_cb)(void *user);

/*
 * @brief       清除用户分组下的所有用户信息   
 * @param[IN]   group_id, 用户组ID
 * @ret         如果删除人脸成功，则返回0，否者返回具体错误码
 */
typedef int (*linkface_clear_usergroup_cb)(const char *group_id);

/*
 * 清空人脸库回调函数，由设备厂商实现
 *
 * */
typedef int (*linkface_clearall_cb)();


/**  当人脸门禁机检测到人脸后，无论是否能正确识别出该人脸，均需要调用本接口将识别结果上报到云端。
 * @param   result 识别结果，如果识别成功，需要包含已识别的人的基本信息。
 *
 * @return 成功返回0，失败返回其他值。
 *
 * 注意:
 * 1. 该接口为同步接口，请勿并发调用。
 * 2. 若接口返回失败，建议最多重试3次。
 */
int linkface_upload_match_result(const linkface_match_result *result);

/* 产品三元组长度限制 */
#define LINKFACE_PRODUCT_KEY_LEN     (20)
#define LINKFACE_DEVICE_NAME_LEN     (32)
#define LINKFACE_DEVICE_SECRET_LEN   (64)

typedef struct {
    char product_key[LINKFACE_PRODUCT_KEY_LEN + 1]; /* 三元组之一，由平台颁发的产品唯一标识，11位长度的英文数字随机组合 */
    char device_name[LINKFACE_DEVICE_NAME_LEN + 1]; /* 三元组之一，在注册设备时，自定义的或自动生成的设备名称，具备产品维度内的唯一性 */
    char device_secret[LINKFACE_DEVICE_SECRET_LEN + 1];  /* 三元组之一，物联网平台为设备颁发的设备密钥*/
} linkface_device_triple_t; /* 物联网平台上分发的三元组信息。 */

/* SDK日志等级 */
typedef enum {
    LF_LOG_DEBUG = 0,
    LF_LOG_INFO = 1,
    LF_LOG_WARN = 2,
    LF_LOG_ERROR = 3,
} linkface_log_level_e;

typedef enum {
    LF_STDOUT = 0,
    LF_FILE = 1
} linkface_log_dest_e;

typedef struct {
    linkface_log_level_e lvl;
    linkface_log_dest_e dst;
} linkface_log_cfg_t;

typedef struct {
    char *logpath;           //日志文件保存路径
    char *picpath;           //人脸照片保存路径
    char *eventpath;         //告警事件文件保存路径
    int max_size_log_mb;    //日志存储目录大小Mb
    int max_size_cache_mb;  //人脸照片存储目录大小Mb
    int max_size_event_mb;  //告警事件存储目录大小Mb
} linkface_cache_cfg_t;

typedef struct {
    int addFaceBatchSize; //每次批量添加人脸的人数，范围为1~15
    linkface_device_triple_t triple;
    linkface_log_cfg_t cfg_log;
    linkface_cache_cfg_t cfg_cache;

    /* 批量增加底库图片*/
    struct {
        linkface_add_image_batch_cb cb;
        void *user;
    } config_batch_add_image;

    /* 删除一个底库图片*/
    struct {
        linkface_del_image_cb cb;
        void *user;
    } config_del_image;

    /* 基于id查询底库图片信息*/
    struct {
        linkface_query_image_info_cb cb;
        void *user;
    } config_query_image_info;

    /* 遍历所有底库图片信息*/
    struct {
        linkface_polling_image_cb cb;
        void *user;
    } config_polling_image_info;

    /* 增加人员信息. */
    struct {
        linkface_add_user_cb cb;
        void *user;
    } config_add_user;

    /* 删除人员信息. */
    struct {
        linkface_del_user_cb cb;
        void *user;
    } config_del_user;

    /* 查询人员信息. */
    struct {
        linkface_query_user_info_cb cb;
        void *user;
    } config_query_user_info;

    /* 轮询所有人员信息. */
    struct {
        linkface_polling_user_info_cb cb;
        void *user;
    } config_polling_user_info;

    /* 聚合接口: 查询端侧人脸总数. */
    struct {
        linkface_querey_image_count_cb cb;
        void *user;
    } config_query_image_count;

    /* 聚合接口:  清空所有人脸和底库图片信息. */
    struct {
        linkface_clearall_cb cb;
    } config_clearall;

    /* 清空指定人脸分组. */
    struct {
        linkface_clear_usergroup_cb cb;
    } config_clear_usergroup;

} linkface_config_t;

/*
 * SDK 初始化接口.
 */
int linkface_init(linkface_config_t *cfg);

/*
 * 消息路由接口，源码开放.
 */
int linkface_message_router(const char *service_name, const int service_name_len, const char *request,
                            const int request_len, char **response, int *response_len);

/** 进程退出前，释放sdk资源。 
 */
void linkface_destroy();

#ifdef __cplusplus
}
#endif

#endif
